@@ -2,11 +2,10 @@ import os
 import json
 from pathlib import Path as P
 from subprocess import call  # nosec
-from typing import Dict, Iterable, List, Optional, Tuple
+from typing import Dict, Iterable, List, Optional, Tuple, Set
 
 import click
 
-from . import git
 from .config import Config
 from .component import Component, component_dir
 from .helpers import relsymlink, kapitan_inventory
@@ -33,14 +32,36 @@ def create_component_symlinks(cfg, component: Component):
         relsymlink(file, cfg.inventory.lib_dir)
 
 
+def _format_component_list(components: Iterable[str]) -> str:
+    formatted_list = list(map(lambda c: f"'{c}'", sorted(components)))
+
+    if len(formatted_list) == 0:
+        return ""
+
+    if len(formatted_list) == 1:
+        return f"{formatted_list[0]}"
+
+    formatted = ", ".join(formatted_list[:-1])
+
+    # Use serial ("Oxford") comma when formatting lists of 3 or more items, cf.
+    # https://en.wikipedia.org/wiki/Serial_comma
+    serial_comma = ""
+    if len(formatted_list) > 2:
+        serial_comma = ","
+
+    formatted += f"{serial_comma} and {formatted_list[-1]}"
+
+    return formatted
+
+
 def _discover_components(cfg) -> Tuple[List[str], Dict[str, str]]:
     """
-    Discover components in `inventory_path/` by extracting all entries from
-    the reclass applications dictionary.
+    Discover components used by the currenct cluster by extracting all entries from the
+    reclass applications dictionary.
     """
     kapitan_applications = kapitan_inventory(cfg, key="applications")
     components = set()
-    component_aliases: Dict[str, str] = {}
+    all_component_aliases: Dict[str, Set[str]] = {}
     for component in kapitan_applications.keys():
         try:
             cn, alias = component.split(" as ")
@@ -53,13 +74,48 @@ def _discover_components(cfg) -> Tuple[List[str], Dict[str, str]]:
                 msg += f" aliased to {alias}"
             click.echo(msg)
         components.add(cn)
+        all_component_aliases.setdefault(alias, set()).add(cn)
 
-        if alias in component_aliases:
-            pc = component_aliases[alias]
-            raise KeyError(
-                f"Duplicate component alias {alias}: component {pc} is already aliased to {alias}"
+    component_aliases: Dict[str, str] = {}
+
+    for alias, cns in all_component_aliases.items():
+        if len(cns) == 0:
+            # NOTE(sg): This should never happen, but we add it for completeness' sake.
+            raise ValueError(
+                f"Discovered component alias '{alias}' with no associated components"
             )
-        component_aliases[alias] = cn
+
+        if len(cns) > 1:
+            if alias in cns:
+                other_aliases = list(cns - set([alias]))
+                if len(other_aliases) > 1:
+                    clist = _format_component_list(other_aliases)
+                    raise KeyError(
+                        f"Components {clist} alias existing component '{alias}'"
+                    )
+
+                # If this assertion fails we have a problem, since `other_aliases` is
+                # the result of removing a single element from a set which contains
+                # multiple elements and we've already handled the case for len() > 1.
+                # Since we don't mind if it's optimized out in some cases, we annotate
+                # it with `nosec` so bandit doesn't complain about it.
+                assert len(other_aliases) == 1  # nosec
+                raise KeyError(
+                    f"Component '{other_aliases[0]}' "
+                    + f"aliases existing component '{alias}'"
+                )
+
+            clist = _format_component_list(cns)
+            raise KeyError(
+                f"Duplicate component alias '{alias}': "
+                + f"components {clist} are aliased to '{alias}'"
+            )
+
+        # len(cns) must be 1 here, as we already raise an Exception for len(cns) ==
+        # 0 earlier. We still assert this condition here and annotate with `nosec`
+        # so bandit doesn't complain about it.
+        assert len(cns) == 1  # nosec
+        component_aliases[alias] = list(cns)[0]
 
     return sorted(components), component_aliases
 
@@ -95,7 +151,13 @@ def _read_components(
         if "version" in info:
             component_versions[component_name] = info["version"]
         else:
-            # Note: We use version=None as a marker for checking out the remote repo's default branch
+            cfg.register_deprecation_notice(
+                f"Component {component_name} doesn't have a version specified. "
+                + "See https://syn.tools/commodore/reference/deprecation-notices.html"
+                + "#_components_without_versions for more details."
+            )
+            # Note: We use version=None as a marker for checking out the remote repo's
+            # default branch.
             component_versions[component_name] = None
         if cfg.debug:
             click.echo(
@@ -131,38 +193,6 @@ def fetch_components(cfg: Config):
         create_component_symlinks(cfg, c)
 
 
-def fetch_jsonnet_libs(config: Config, libs):
-    """
-    Download all libraries specified in list `libs`.
-    Each entry in `libs` is assumed to be a dict with keys
-      * 'repository', the value of which is interpreted as a git repository to
-                      clone.
-      * 'files', a list of dicts which defines which files in the repository
-                 should be installed as template libraries.
-    Each entry in files is assumed to have the keys
-      * 'libfile', indicating a filename relative to the repository of the
-                   library to install
-      * 'targetfile', the file name to use as the symlink target when
-                      installing the library
-    """
-
-    click.secho("Updating Jsonnet libraries...", bold=True)
-    config.inventory.ensure_dirs()
-    for lib in libs:
-        libname = lib["name"]
-        if config.debug:
-            filestext = " ".join([f["targetfile"] for f in lib["files"]])
-            click.echo(f" > {libname}: {filestext}")
-        library_dir = config.inventory.libs_dir / libname
-        git.clone_repository(lib["repository"], library_dir, config)
-        for file in lib["files"]:
-            relsymlink(
-                library_dir / file["libfile"],
-                config.inventory.lib_dir,
-                dest_name=file["targetfile"],
-            )
-
-
 def jsonnet_dependencies(config: Config) -> Iterable:
     """
     Creates a list of Jsonnet dependencies for the given Components.
@@ -210,6 +240,7 @@ def write_jsonnetfile(file: P, deps: Iterable):
 
     with open(file, "w", encoding="utf-8") as f:
         f.write(json.dumps(data, indent=4))
+        f.write("\n")
 
 
 def fetch_jsonnet_libraries(cwd: P, deps: Iterable = None):
@@ -270,6 +301,7 @@ def inject_essential_libraries(file: P):
 
     with open(file, "w", encoding="utf-8") as j:
         json.dump(data, j, indent=4)
+        j.write("\n")
 
 
 def register_components(cfg: Config):
@@ -280,7 +312,10 @@ def register_components(cfg: Config):
     Create component symlinks for discovered components which exist.
     """
     click.secho("Discovering included components...", bold=True)
-    components, component_aliases = _discover_components(cfg)
+    try:
+        components, component_aliases = _discover_components(cfg)
+    except KeyError as e:
+        raise click.ClickException(f"While discovering components: {e}")
     click.secho("Registering components and aliases...", bold=True)
 
     for cn in components:
